import numpy as np
import pandas as pd
from .plotting import threshold_plot

def hotelling_t2(df, num_in_control, alpha=.01, plotting=True, save="", plot_title="Hotellings"):
	"""
	Args:
		df: multivariate dataset as Pandas DataFrame
		num_in_control: number of in control observations before the anomalies start
		alpha: the percentage of false positives we want to allow, used for calculating the Upper Control Limit (default .01 or 1%)
			save: the directory to save the graphs to, if not changed from default, nothing will be saved
			plot_title: the title for the plot generated
	Returns:
		t^2 statistic values and the UCL
	"""
	# calculate covariance matrix
	S = df.head(num_in_control).cov()

	# finding x-mean vector for each row in dataframe
	s_inv = pd.DataFrame(np.linalg.pinv(S), S.columns, S.index)
	mean_vec = df.head(num_in_control).mean()  # 56 values
	t2_values = []
	for index, row in df.iterrows():
		diff = row.subtract(mean_vec)
		# calculate T2 values now: T2 = [x-mean]'S-1[x-mean]
		t2 = diff.transpose() @ s_inv
		t2 = t2@diff
		t2_values.append(t2)
	   
	#calculate UCL
	in_stats = t2_vals[0:num_in_control]
	ucl = max(in_stats)
 
	count = len([i for i in in_stats if i > ucl]) 

	while(count < (alpha* len(in_stats))):
			ucl = ucl - step_size
			count = len([i for i in in_stats if i > ucl])  
	  
	#plotting
	if plotting:
		fig, ax = plt.subplots(figsize=(10, 7))
		lc = threshold_plot(ax, range(0, df.shape[0]), array(t2_vals), ucl, 'b', 'r')
		ax.axhline(ucl, color='k', ls='--')
		ax.set_title(plot_title)
		ax.set_xlabel('Observation Number')
		ax.set_ylabel('Hotelling statistic (Anomaly Score)')

		#saving plot
		if save != "":
			if save[-1] != '/':
			save += '/'
			plt.savefig(save + plot_title + '_FULL.png', dpi=300)

			#if there are a lot of entries, a smaller graph will also be saved that
			#shows only the end of the graph for more detail.
			if len(y_vals) > 10000:
			ax.set_xlim(9000)
			plt.savefig(save + plot_title + '_SMALL.png', dpi=300)

	# return t2 values
	return t2_values, ucl



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="mitten",
	version="0.0.9a",
	author="Marco Christiani",
	author_email="mchristiani2017@gmail.com",
	description="A package for multivariate statistical process control modeling.",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/Marco-Christiani/MITTEN",
	packages=setuptools.find_packages(exclude=["tests"]),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Operating System :: OS Independent",
	],
	python_requires='>=3.6',
	install_requires=[
		'scipy',
		'numpy',
		'pandas',
		'matplotlib',
		'pyplot-themes',
		'seaborn'
	]
)


import pandas as pd
import numpy as np

def _univariate_t_test(data, n_in_control, alpha=0.05):
	"""
	Diagnostic test to be used as a followup to a multivariate signal
	t=(mu_new-mu-ref)/sqrt(sd_ref*(1/n_new+1/n_ref))
	Where ref represents in control data and new represents potetially out of control data

	Args:
		data: data used to calculate t-statistic
		n_in_control: the first `n_in_control` data points of `data` are sample1, the remaining are sample2
		alpha: significance level of the test (default = 5%)
	Returns:
		Calculated t-statistic for the test data[:n_in_control] vs data[n_in_control:]
	"""
	ref_data = data[:n_in_control]
	new_data = data[n_in_control:]
	n_ref = len(ref_data)
	n_new = len(new_data)

	mean_ref = ref_data.mean()
	sd_ref = ref_data.std()

	mean_new = new_data.mean()

	t = (mean_new - mean_ref) / np.sqrt(sd_ref * (1 / n_new + 1 / n_ref))
	return t


def _build_t_test_df(df, in_control_start, batch_size):
	"""
	Constructs a dataframe containing t-test statistics from multiple t-tests, the number
	of which is controlled by batch_size

	Args:
		df: dataframe which stores the data to be tested, with features as columns 
			and observations as rows.
		in_control_start: the starting index of the in control data
		batch_size: successive t-tests will be run on subsets of the potentially out of control
			segment of the dataset contained in `df`. The size of each subset (aka batch) is 
			controlled by `batch_size`
	Returns:
		Pandas DataFrame with floor(n_out_control/batch_size) rows where n_out_control=(len(df)-in_control_start).
		Each cell stores a test statistic.
	"""
	t_df = pd.DataFrame(columns=df.columns)
	# t_matrix = []
	m = 0
	for in_control in range(in_control_start, len(df), batch_size):
		row = {}
		for col in df.columns:
			end_index = in_control + batch_size
			t = _univariate_t_test(df[col][:end_index], in_control, alpha=0.05)
			row[col] = t
		t_df = t_df.append(row, ignore_index=True)
		m += 1
	t_df.index = list(range(in_control_start, len(df), batch_size))
	return t_df


def interpret_multivariate_signal(df,
								  n_in_control,
								  batch_size=5,
								  n_most_likely=5,
								  verbose=False):
	"""
	Designed to interpret an out of control signal from a multivariate control chart method. Used
	to identify the source of the signal by examining each individual feature using successive t-tests.
	Since this method relies on t-tests, it is ill-equipped to handle shifts in process variability.

	Args:
		df: dataframe which stores the data to be tested, with features as columns 
			and observations as rows.
		n_in_control: the number of observations which are in control (the observations before the signal)
		batch_size: successive t-tests will be run on subsets of the potentially out of control
			segment of the dataset contained in `df`. The size of each subset (aka batch) is 
			controlled by `batch_size`
		n_most_likely: if `verbose=True` then this controls how many of the most likely observations will
			be printed.
		verbose: if `True`, prints the most likely cuplrit features
	Returns:
		A ranked list of features (as a Pandas series) sorted from highest to lowest average t-statistic ranking.
	"""
	t_test_df = _build_t_test_df(df, n_in_control, batch_size)

	ranks_srs = pd.Series(index=df.columns)
	ranks_srs[ranks_srs.index] = 0
	for index, row in t_test_df.iterrows():
		t_stats = row.sort_values(ascending=False) # bug? these should be magnitude-ranked
		ranks_srs[t_stats.index] += row.rank(ascending=False) # Rank each t-statistic
	ranks_srs = ranks_srs / len(
		t_test_df)  # Convert t statistics to an average ranking
	if verbose:
		print(
			'The most likely culprit features and average t-statistic ranking in decreasing order are:'
		)
		print(ranks_srs.sort_values()[:n_most_likely])
	return ranks_srs.sort_values()


import numpy as np
from numpy import matmul, sqrt, array
from .plotting import threshold_plot

def mcusum(df, num_in_control,k,alpha=.01, plotting=True, save="", plot_title="MCUSUM"):
	"""
	Implementation of the Multivariate Cumulative Sum (MCUSUM) method.

		Based on Kent(2007) https://etd.ohiolink.edu/rws_etd/send_file/send?accession=kent1185558637&disposition=inline

		- Reference 17 : (Jackson 1985)
		
		- Reference 5  : (Crosier 1988)

	Args:
		df: multivariate dataset as Pandas DataFrame
		num_in_control: number of in control observations
		k: the slack parameter which determines model sensetivity (should typically be set to 1/2 of the mean shift that you expect to detect)
		alpha: the percentage of false positives we want to allow, used for calculating the Upper Control Limit (default .01 or 1%)
		save: the directory to save the graphs to, if not changed from default, nothing will be saved
		plot_title: the title for the plot generated
	Returns:
		MCUSUM statistic values as a list
	"""
	a = df.head(num_in_control).mean(axis=0)  # mean vector of in control data
	cov_inv = np.linalg.inv(np.cov(df.head(num_in_control), rowvar=False,bias=True))  # covariance matrix inverted
	s_old = [0] * df.shape[1]
	y_vals = [0] * df.shape[0]
	for n in range(0, df.shape[0]):
		# get current line
		x = df.iloc[n]

		# calculate Cn
		tf = array(s_old + x - a)
		c = matmul(matmul(tf, cov_inv), tf)
		c = sqrt(c)

		# calculate kv (vector k)
		kv = (k / c) * tf

		# calculate new s
		if (c <= k):
			s_old = 0
		else:
			s_old = tf * (1 - (k / c))
			# calculate y (new c)
			tf = array(s_old + x - a)
			y = matmul(matmul(tf, cov_inv), tf)
			y = sqrt(y)
			y_vals[n] = y
	
	#calculate UCL
	in_stats = y_vals[0:num_in_control]
	ucl = max(in_stats)
 
	count = len([i for i in in_stats if i > ucl]) 

	while(count < (alpha* len(in_stats))):
			ucl = ucl - step_size
			count = len([i for i in in_stats if i > ucl])
	
	#plotting
	if plotting:
		fig, ax = plt.subplots(figsize=(10, 7))
		lc = threshold_plot(ax, range(0, df.shape[0]), array(y_vals), ucl, 'b', 'r')
		ax.axhline(ucl, color='k', ls='--')
		ax.set_title(plot_title)
		ax.set_xlabel('Observation Number')
		ax.set_ylabel('MCUSUM statistic (Anomaly Score)')

		#saving plot
		if save != "":
			if save[-1] != '/':
				save += '/'
			plt.savefig(save + plot_title + '_FULL.png', dpi=300)
			
			#if there are a lot of entries, a smaller graph will also be saved that
			#shows only the end of the graph for more detail.
			if len(y_vals) > 10000:
				ax.set_xlim(9000)
				plt.savefig(save + plot_title + '_SMALL.png', dpi=300)
			
		
	#returns the mcusum statistics as a list and upper control limit
	return y_vals, ucl



import numpy as np
import pyplot_themes as themes
from matplotlib.collections import LineCollection
from matplotlib.colors import ListedColormap, BoundaryNorm


def threshold_plot(ax, x, y, threshv, color, overcolor):
	"""
	Helper function to plot points above a threshold in a different color
	From https://stackoverflow.com/questions/30121773/python-is-it-possible-to-change-line-color-in-a-plot-if-exceeds-a-specific-range

	Args:
		ax : Axes
			Axes to plot to
		x, y : array
			The x and y values

		threshv : float
			Plot using overcolor above this value

		color : color
			The color to use for the lower values

		overcolor: color
			The color to use for values over threshv
	"""
	themes.theme_ggplot2()
	# import numpy as np
	# Create a colormap for red, green and blue and a norm to color
	# f' < -0.5 red, f' > 0.5 blue, and the rest green
	cmap = ListedColormap([color, overcolor])
	norm = BoundaryNorm([np.min(y), threshv, np.max(y)], cmap.N)

	# Create a set of line segments so that we can color them individually
	# This creates the points as a N x 1 x 2 array so that we can stack points
	# together easily to get the segments. The segments array for line collection
	# needs to be numlines x points per line x 2 (x and y)
	points = np.array([x, y]).T.reshape(-1, 1, 2)
	segments = np.concatenate([points[:-1], points[1:]], axis=1)

	# Create the line collection object, setting the colormapping parameters.
	# Have to set the actual values used for colormapping separately.
	lc = LineCollection(segments, cmap=cmap, norm=norm)
	lc.set_array(y)

	ax.add_collection(lc)
	ax.set_xlim(np.min(x), np.max(x))
	ax.set_ylim(np.min(y) * 1.1, np.max(y) * 1.1)
	return lc



import numpy as np
import pandas as pd
from numpy.linalg import inv as inverse
from .plotting import threshold_plot


def apply_mewma(df, num_in_control, lambd=0.1, alpha=.01, plotting=True, save="", plot_title="MEWMA"):
	"""
	Args:
		df: multivariate dataset as Pandas DataFrame
		num_in_control: number of rows before anomalies begin
		lambd: smoothing parameter between 0 and 1; lower value = higher weightage to older observations; default is 0.1
		alpha: the percentage of false positives we want to allow, used for calculating the Upper Control Limit (default .01 or 1%)
		save: the directory to save the graphs to, if not changed from default, nothing will be saved
		plot_title: the title for the plot generated
	Returns:
		MEWMA statistic values
	"""
	nrow, ncol = df.shape
	means = df.head(num_in_control).mean(axis=0)

	# create diff matrix
	v = np.zeros(shape=(nrow - 1, ncol))
	for i in range(nrow - 1):
		v[i] = df.iloc[i + 1] - df.iloc[i]

	# calculate vTv
	vtv = v.T @ v

	# calculate S matrix
	S = (1 / (2 * (nrow-1))) * (vtv)

	mx = df - means

	# calculate z
	z = np.zeros(shape=(nrow + 1, ncol))
	for i in range(nrow):
		z[i + 1] = lambd * mx.iloc[i] + (1 - lambd) * z[i]
	z = z[1:, :]

	t2 = []  # test statistic values
	for i in range(nrow):
		w = (lambd / (2 - lambd)) * (1 - (1 - lambd)**(2 * (i + 1)))
		inv = inverse(w * S)
		t2.append((z[i].T @ inv) @ z[i])
		
	#calculate UCL
	in_stats = t2[0:num_in_control]
	ucl = max(in_stats)
 
	count = len([i for i in in_stats if i > ucl]) 

	while(count < (alpha* len(in_stats))):
			ucl = ucl - step_size
			count = len([i for i in in_stats if i > ucl])
	
	#plotting
	if plotting:
		fig, ax = plt.subplots(figsize=(10, 7))
		lc = threshold_plot(ax, range(0, df.shape[0]), array(t2), ucl, 'b', 'r')
		ax.axhline(ucl, color='k', ls='--')
		ax.set_title(plot_title)
		ax.set_xlabel('Observation Number')
		ax.set_ylabel('MEWMA statistic (Anomaly Score)')

		#saving plot
		if save != "":
			if save[-1] != '/':
				save += '/'
			plt.savefig(save + plot_title + '_FULL.png', dpi=300)
			
			#if there are a lot of entries, a smaller graph will also be saved that
			#shows only the end of the graph for more detail.
			if len(y_vals) > 10000:
				ax.set_xlim(9000)
				plt.savefig(save + plot_title + '_SMALL.png', dpi=300)

	# return t2 values
	return t2


def pc_mewma(df, num_in_control, num_princ_comps, alpha=.01, plotting=True, save="", plot_title="PC_MEWMA"):
	"""
	MEWMA on Principle Components
	Variables contained in `df` must have mean 0

	Args:
		df: multivariate dataset as Pandas DataFrame
		num_in_control: number of in control observations
		num_princ_comps: number of principle components to include
		alpha: the percentage of false positives we want to allow, used for calculating the Upper Control Limit (default .01 or 1%)
		save: the directory to save the graphs to, if not changed from default, nothing will be saved
		plot_title: the title for the plot generated
	Returns:
		MEWMA statistic values using PCA for dimensionality reduction
	"""
	df = df-df.mean()
	pca = PCA(n_components=num_princ_comps)
	s = pca.fit(df)
	eigvec_mat = s.components_ # V transpose
	W_matrix = []  # since apply_mewma only takes a df, not observations as they happen
	for index, row in df.iterrows():
		W = eigvec_mat @ row
		W_matrix.append(W)
	W_df = pd.DataFrame(W_matrix)
	return apply_mewma(W_df,
					   num_in_control,
					   lambd=0.1,
					   alpha,
					   plotting,
					   save,
					   plot_title)
